/*div.title-bar{
  margin-bottom: 5px;
  border-bottom: solid;
  border-color: gray;
}
.title-bar .col-12{
  background: #ccc;
  margin-top: 5px;
}

div.title-bar img {
  padding-top: 5px;
  max-height: 55px;
  min-height: 40px;
  height: calc( 9vh - 15px);
}*/
div.container{
  margin-top: 30px;
}
h1.page-title{
  font-size: 240%;
  font-family: Garamond, Georgia, serif;
  font-weight: normal;
  margin-bottom: 10px;
  padding: 0px;
}

  