div.detail-pane {
  height: 90vh;
  overflow-y: scroll;
  overflow-x: hidden;
  border-style: solid;
  border-width: 1px;
  padding: 5px;
  padding-left: 10px;
}

div.group {
  border-top-width: 2px;
  border-color: #d4d1d1;
  border-top-style: solid;
}
div.group > h4 {
  font-size: 1.2em;
  font-weight: 100;
  background: #f5f3f3;
  padding-left: 10px;
  padding-top: 10px;
  padding-bottom: 10px;
}
::ng-deep th {
  background: #f6f9fb 
}
pre.blue-stripes {
  position: relative;
  border-left: 10px solid #358ccb;
  box-shadow: -1px 0px 0px 0px #358ccb, 0px 0px 0px 1px #dfdfdf;
  background-color: #fdfdfd;
  background-image: linear-gradient(transparent 50%, rgba(69, 142, 209, 0.04) 50%);
  background-size: 3em 3em;
  background-origin: content-box;
  background-attachment: local;
}
pre.blue-stripes code{
  background: inherit;
  border-radius: 0px;
  border-style: inherit;

}
::ng-deep pre code{
  border-radius: 4px;
  border-style: solid;
  border-radius: 4px;
  border-width: 1px;
  border-color: gray;
  background: #e2f5f5;
  
}

pre[class*="language-"]:before,
pre[class*="language-"]:after {
    display: none;
}
