.header.ini {
  margin-left: 20px;
  margin-right: 30px;
  padding-left: 15px;
  font-family: monospace;
  font-size: 15px;
  color: black;
  border: 1px;
  border-style: solid;
  border-top-left-radius: 3px;
  border-top-right-radius: 3px;
  border-bottom: 0px;
  background: #ececec;
  padding: 8px;
  color: #636262;
}

.hjs-code {
  border-style: solid;
  border-width: 1px;
  padding: 15px;
  margin-left: 20px;
  margin-right: 30px;
  margin-bottom: 15px;
  white-space: pre;
}
.hjs-code {
  padding: 0px;
}
.hjs-code > pre > code.hljs {
  border-width: 0px;
}

.hjs-code > pre {
  margin: 0px;
}

.code.ini {
  border-style: solid;
  border-width: 1px;
  padding: 15px;
  margin-left: 20px;
  margin-right: 30px;
    margin-bottom: 15px;
  white-space: pre;
  font-family: monospace;
  font-size: 14px;
  border-color: black;
}

.ini>b {
  color: #1b80e4;
}

.ini>span {
  color: #007021;
  font-weight: bold;
}

.ini>i {
  color: gray;
}

.header.bash {
  border: 1px;
  border-style: solid;
  border-bottom: 0;
  border-color: grey;
  padding-left: 10px;
  background: grey;
  color: white;
}

.code.bash {
  background: #545454;
  border: 2px;
  border-style: solid;
  border-color: grey;
  padding: 5px;
  font-family: monospace;
  font-size: 14px;
  color: #fff;
  white-space: pre;
  line-height: normal;
  margin-bottom: 20px;
  overflow-x: auto;
}
code {
  white-space: pre !important;
}

.bash>i {
  color: #dbdbdb;
}

.bash>p>i {
  color: #dbdbdb;
}

.bash>p {
  margin-top: 0px;
}

.command-title {
  background: #DDD;
  padding: 8px;
  border-radius: 4px;
  margin-top: 40px;
  border-top: black;
  border-style: solid;
  border-width: 1px
}

.command-title>i {
  font-style: normal;
  font-weight: bold;
  font-family: monospace;
}

.text-mono {
  font-family: monospace;
}

.block-indent {
  margin-left: 20px;
}

.text-grey {
  color: #808080;
}

.arglist>dd {
  margin: 0px;
  margin-left: 30px;
}

.arglist>dt {
  color: #808080;
  font-weight: normal;
  font-family: monospace;
}

.arglist {
  margin-top: 0px;
  margin-bottom: 0px;
}

.plainlist {
  margin-bottom: 0px;
}

.plainlist>dd {
  margin-left: 30px;
}

.plainlist>dt {
  font-weight: normal;
  font-family: monospace;
}

.plug-cli {
  font-family: monospace;
  background: #DDD;
  display: block;
  margin: 5px;
  padding: 5px;
  border-radius: 4px;
  border-style: solid;
  border-width: 1px;
  border-color: #BBB
}

.mono-dl>dt {
  font-family: monospace;
}

.arglist.dl-horizontal>dt {
  width: inherit;
}

.arglist.dl-horizontal>dd {
  margin-left: 130px;
}

.command {
  font-family: monospace;
  color: #808080
}

.command>em {
  font-weight: bold;
  font-style: normal;
  color: black;
}

/* styles for CLI page */

#command-line-interface h3{
  display: inline;
  font-family: monospace;
  font-size: 1.6rem;
  font-weight: 200;
}

#command-line-interface h3 + p {
  display: inline;
  font-style: italic;
}

#command-line-interface .docutils dl{
    font-weight: 100;
}

#command-line-interface .docutils dt{
    margin-bottom: 0px;
    font-weight: 100;
    margin-top: 5px;
}

#command-line-interface .docutils dd{
    margin-bottom: 0px;
}

.cli-hr{
    margin-top: 0px;
    margin-bottom: 8px;
}
.cli-cmd{
    font-size: 1.2rem;
    font-family: monospace;
    font-weight: bold;
    display: inline;
    margin-bottom: 10px;
}
.cli-args{
    font-size: 1.2rem;
    color: gray;
    font-family: monospace;
}
.arg-div{
    display: flex;
}
.arg-dt{
    font-family: monospace;
    font-size: 1.2rem;
    color: gray;
    min-width: 140px;
}
.arg-dd{
    flex-grow: 4;
    margin-left: 10px;

}